#[allow(non_upper_case_globals)]
#[allow(non_camel_case_types)]
#[allow(non_snake_case)]
#[allow(dead_code)]
mod lean {
	include!(concat!(env!("OUT_DIR"), "/lean.rs"));

	#[link(name = "leanshared")]
	extern "C" {
		pub fn lean_initialize_runtime_module();
		pub fn lean_initialize();
		// lean_io_mark_end_initialization is already defined.
	}
}

mod callee {
	use super::lean::*;

	#[link(name = "Callee")]
	extern "C" {
		#[link_name = "initialize_Callee"]
		pub fn initialize(builtin: u8, world: lean_obj_arg) -> lean_obj_res;
		pub fn my_function(s: lean_obj_arg) -> lean_obj_res;
		pub fn stylize(s: lean_obj_arg) -> lean_obj_res;
		pub fn think(s: lean_obj_arg) -> lean_obj_res;
		pub fn print_in_upper(s: lean_obj_arg, world: lean_obj_arg) -> lean_obj_res;
	}
}

fn initialize() {
	unsafe {
		lean::lean_initialize_runtime_module();
		lean::lean_initialize();
		let builtin: u8 = 1;
		let res = callee::initialize(builtin, lean::lean_io_mk_world());
		if lean::lean_io_result_is_ok(res) {
			lean::lean_dec_ref(res);
		} else {
			lean::lean_io_result_show_error(res);
			lean::lean_dec(res);
			panic!("Failed to load callee");
		}
		lean::lean_io_mark_end_initialization();
	}
}

fn call_my_function() {
	let c_str = std::ffi::CString::new("hello").unwrap();
	let result = unsafe {
		let native_str = lean::lean_mk_string(c_str.as_ptr());
		let result_obj = callee::my_function(native_str);
		let result_c_str = std::ffi::CStr::from_ptr(lean::lean_string_cstr(result_obj));
		let str = result_c_str.to_str().unwrap().to_string();
		// Clean up the resultant objects
		lean::lean_dec(result_obj);
		str
	};
	assert_eq!(result, "hello at callee");
}

fn unbox_lean_string(str: lean::b_lean_obj_arg) -> String {
	unsafe { std::ffi::CStr::from_ptr(lean::lean_string_cstr(str)) }
		.to_str()
		.expect("Failed to convert C-string to string")
		.to_string()
}

fn call_stylize() {
	let input = std::ffi::CString::new("aBcDe").unwrap();
	let (s1, s2, s3) = unsafe {
		let result_obj = callee::stylize(lean::lean_mk_string(input.as_ptr()));
		assert!(lean::lean_is_ctor(result_obj), "Tuple must be a ctor");
		// Remember Lean tuples are nested!
		assert_eq!(
			lean::lean_ctor_num_objs(result_obj),
			2,
			"Tuple ctor has 3 elements"
		);
		let s1 = unbox_lean_string(lean::lean_ctor_get(result_obj, 0));
		let result_obj2 = lean::lean_ctor_get(result_obj, 1);
		let s2 = unbox_lean_string(lean::lean_ctor_get(result_obj2, 0));
		let s3 = unbox_lean_string(lean::lean_ctor_get(result_obj2, 1));
		lean::lean_dec(result_obj);
		(s1, s2, s3)
	};
	assert_eq!(s1, "ABcDe");
	assert_eq!(s2, "aBcDe");
	assert_eq!(s3, "ABCDE");
}

fn call_think(s: &str) -> Result<String, usize> {
	let input = std::ffi::CString::new(s).expect("Failed to convert to C-string");
	unsafe {
		let result_obj = callee::think(lean::lean_mk_string(input.as_ptr()));
		let tag = lean::lean_ptr_tag(result_obj);
		let result = if tag == 0 {
			assert!(lean::lean_is_ctor(result_obj), "Think must be a ctor");
			assert_eq!(
				lean::lean_ctor_num_objs(result_obj),
				1,
				"Ctor must have 1 field"
			);
			let s = unbox_lean_string(lean::lean_ctor_get(result_obj, 0));
			Ok(s)
		} else if tag == 1 {
			assert_eq!(
				lean::lean_ctor_num_objs(result_obj),
				1,
				"Ctor must have 1 field"
			);
			let field = lean::lean_ctor_get(result_obj, 0);
			let n = lean::lean_uint64_of_nat(field) as usize;
			Err(n)
		} else {
			unreachable!("Invalid constructor id")
		};
		lean::lean_dec(result_obj);
		result
	}
}

fn call_print_in_upper()
{
	let c_str = std::ffi::CString::new("caller").unwrap();
	unsafe {
		let native_str = lean::lean_mk_string(c_str.as_ptr());
		let res = callee::print_in_upper(native_str, lean::lean_io_mk_world());
		if lean::lean_io_result_is_ok(res) {
			lean::lean_dec_ref(res);
		} else {
			lean::lean_io_result_show_error(res);
			lean::lean_dec(res);
			panic!("IO Monad execution failed");
		}
	};
}

fn main() {
	initialize();
	println!("Lean initialization complete!");

	call_my_function();
	call_stylize();

	assert_eq!(call_think("Stanford"), Ok("tanford".to_string()));
	assert_eq!(call_think("Else"), Err(4));
	call_print_in_upper();
}
